#include <config.h>

#include <iostream>

#include <dune/common/bitsetvector.hh>

#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/operators.hh>
#include <dune/istl/solvers.hh>
#include <dune/istl/solver.hh>
#include <dune/istl/preconditioners.hh>

#include <dune/solvers/norms/energynorm.hh>
#include <dune/solvers/solvers/loopsolver.hh>
#include <dune/solvers/iterationsteps/cgstep.hh>
#include <dune/solvers/solvers/cgsolver.hh>


#include "common.hh"

/**  \brief test for the CGStep class
  *
  *  This test tests if the Dirichlet problem for a Laplace operator
  *  is solved correctly for a random rhs by a LoopSolver employing
  *  the CGStep.
  */
struct CGTestSuite
{
    template <class GridType>
    bool check(const GridType& grid)
    {
        double stepTol    = 1e-12; // termination criterion
        double solveTol   = 1e-6;  // error in the solution
        int maxIterations = 1000;

        bool passed = true;

        static const int dim = GridType::dimension;

        typedef typename Dune::FieldMatrix<double, 1, 1> MatrixBlock;
        typedef typename Dune::FieldVector<double, 1> VectorBlock;
        typedef typename Dune::BCRSMatrix<MatrixBlock> Matrix;
        typedef typename Dune::BlockVector<VectorBlock> Vector;
        typedef typename Dune::BitSetVector<1> BitVector;

        typedef typename GridType::LevelGridView GridView;

        const GridView gridView = grid.levelView(grid.maxLevel());

        Matrix A;
        constructPQ1Pattern(gridView, A);
        A=0.0;
        assemblePQ1Stiffness(gridView, A);
        EnergyNorm<Matrix, Vector> energyNorm(A);

        Vector u(A.N()), u_sol(A.N());
        u = 0;
        for (size_t i=0; i<u.size(); ++i)
            u_sol[i] = (1.0*rand())/RAND_MAX;

        Vector rhs(A.N());
        A.mv(u_sol, rhs);

        BitVector ignore(A.N());
        ignore.unsetAll();
        markBoundaryDOFs(gridView, ignore);

        // set Dirichlet values
        for (size_t i=0; i<u.size(); ++i)
            if (ignore[i][0])
                u[i] = u_sol[i];

        Vector u1 = u;
        Vector u2 = u;
        Vector u3 = u;

        Vector rhs1 = rhs;
        Vector rhs2 = rhs;
        Vector rhs3 = rhs;

        // Test 1: Dune::CGSolver (dune-istl)
        {
            Dune::MatrixAdapter<Matrix,Vector,Vector> op(A);
            Dune::Richardson<Vector,Vector> richardson;
            Dune::CGSolver<Vector> solver(op, richardson, stepTol, maxIterations, 0); // verbosity
            Dune::InverseOperatorResult statistics;
            solver.apply(u1, rhs1, statistics);

            const auto normDiff1 = energyNorm.diff(u_sol, u1);
            std::cout << "Energy norm of the error for Dune::CGSolver (dune-istl):           "
                      << normDiff1 << std::endl;
            if (normDiff1 > solveTol) {
                std::cerr << "Error too large for Dune::CGSolver (dune-istl)" << std::endl;
                passed = false;
            }
        }

        // Test 2: CGSolver (dune-solvers)
        {
            CGSolver<Matrix,Vector> solver(&A, &u2, &rhs2, nullptr,
                                           maxIterations, stepTol, &energyNorm,
                                           Solver::QUIET, false); // absolute error
            solver.check();
            solver.preprocess();
            solver.solve();

            const auto normDiff2 = energyNorm.diff(u_sol, u2);
            std::cout << "Energy norm of the error for CGSolver (dune-solvers):              "
                      << normDiff2 << std::endl;
            if (normDiff2>solveTol) {
                std::cerr << "Error too large for Dune::CGSolver (dune-solvers)" << std::endl;
                passed = false;
            }
        }

        // Test 3: Dune::Solvers::CGStep (dune-solvers)
        {
            Dune::Solvers::CGStep<Matrix,Vector> cgStep(A, u3, rhs3);
            LoopSolver<Vector> solver(&cgStep, maxIterations, stepTol, &energyNorm,
                                      Solver::QUIET, false); // absolute error
            solver.check();
            solver.preprocess();
            solver.solve();

            const auto normDiff3 = energyNorm.diff(u_sol, u3);
            std::cout << "Energy norm of the error for Dune::Solvers::CGStep (dune-solvers): "
                      << normDiff3 << std::endl;
            if (normDiff3>solveTol) {
                std::cerr << "Error too large for Dune::Solvers::CGStep (dune-solvers)" << std::endl;
                passed = false;
            }
        }
        return passed;
    }
};

int main(int argc, char** argv) try
{
    bool passed(true);

    CGTestSuite testsuite;
    passed = checkWithStandardGrids(testsuite);

    return passed ? 0 : 1;
}
catch (Dune::Exception e) {

    std::cout << e << std::endl;

}

