#ifndef DUNE_SOLVERS_TESTS_COMMON_HH
#define DUNE_SOLVERS_TESTS_COMMON_HH


#include <dune/common/fvector.hh>
#include <dune/common/fmatrix.hh>

#include <dune/istl/matrixindexset.hh>

#include <dune/geometry/quadraturerules.hh>

#include <dune/localfunctions/lagrange/pqkfactory.hh>

#include <dune/grid/common/gridfactory.hh>

#if HAVE_UG
#include <dune/grid/uggrid.hh>
#endif

#if HAVE_ALUGRID
#include <dune/grid/alugrid.hh>
#endif

#if HAVE_DUNE_SUBGRID
#include "dune/subgrid/subgrid.hh"
#endif


template<class GridView, class Matrix>
void constructPQ1Pattern(const GridView& gridView, Matrix& matrix)
{
    static const int dim = GridView::Grid::dimension;

    typedef typename GridView::IndexSet IndexSet;
    typedef typename GridView::template Codim<0>::Iterator ElementIterator;
    typedef typename GridView::template Codim<0>::Entity Element;
    typedef typename Dune::PQkLocalFiniteElementCache<double, double, dim, 1> FiniteElementCache;
    typedef typename FiniteElementCache::FiniteElementType FiniteElement;

    const IndexSet& indexSet = gridView.indexSet();
    FiniteElementCache cache;

    int size = indexSet.size(dim);

    Dune::MatrixIndexSet indices(size, size);

    ElementIterator it = gridView.template begin<0>();
    ElementIterator end = gridView.template end<0>();
    for (; it != end; ++it)
    {
        Element& e = *it;
        const FiniteElement& fe = cache.get(e.type());

        int localSize = fe.localBasis().size();
        for (int i = 0; i < localSize; ++i)
        {
            int iGlobal = indexSet.subIndex(e, fe.localCoefficients().localKey(i).subEntity(), dim);
            for (int j = 0; j < localSize; ++j)
            {
                int jGlobal = indexSet.subIndex(e, fe.localCoefficients().localKey(j).subEntity(), dim);
                indices.add(iGlobal, jGlobal);
                indices.add(jGlobal, iGlobal);
            }
        }
    }
    indices.exportIdx(matrix);
}


template<class GridView, class Matrix>
void assemblePQ1Stiffness(const GridView& gridView, Matrix& matrix)
{
    static const int dim = GridView::Grid::dimension;
    static const int dimworld = GridView::Grid::dimensionworld;

    typedef typename GridView::IndexSet IndexSet;
    typedef typename GridView::template Codim<0>::Iterator ElementIterator;
    typedef typename GridView::template Codim<0>::Entity Element;
    typedef typename GridView::template Codim<0>::Entity::Geometry Geometry;
    typedef typename Dune::PQkLocalFiniteElementCache<double, double, dim, 1> FiniteElementCache;
    typedef typename FiniteElementCache::FiniteElementType FiniteElement;

    typedef typename Dune::QuadratureRule<double,dim> QuadratureRule;
    typedef typename Dune::template FieldVector<double,dim> LocalCoordinate;
    typedef typename Dune::template FieldVector<double,dimworld> GlobalCoordinate;
    typedef typename Element::Geometry::JacobianInverseTransposed JacobianInverseTransposed;
    typedef typename FiniteElement::Traits::LocalBasisType::Traits::JacobianType JacobianType;

    const IndexSet& indexSet = gridView.indexSet();
    FiniteElementCache cache;

    ElementIterator it = gridView.template begin<0>();
    ElementIterator end = gridView.template end<0>();
    for (; it != end; ++it)
    {
        Element& e = *it;
        const Geometry geometry = e.geometry();
        const FiniteElement& fe = cache.get(e.type());

        int localSize = fe.localBasis().size();

        // get quadrature rule of appropiate order (P1/Q1)
        int order = 0;
        if (e.type().isSimplex())
            order = 2*(1-1);
        else
            order = 2*(dim-1);
        const QuadratureRule& quad = Dune::template QuadratureRules<double, dim>::rule(e.type(), order);

        // store gradients of shape functions and base functions
        std::vector<JacobianType> referenceGradients(localSize);
        std::vector<GlobalCoordinate> gradients(localSize);

        for (size_t pt=0; pt < quad.size(); ++pt)
        {
            // get quadrature point
            const LocalCoordinate& quadPos = quad[pt].position();

            // get transposed inverse of Jacobian of transformation
            const JacobianInverseTransposed invJacobian = geometry.jacobianInverseTransposed(quadPos);

            // get integration factor
            const double integrationElement = geometry.integrationElement(quadPos);

            // evaluate gradients of shape functions
            fe.localBasis().evaluateJacobian(quadPos, referenceGradients);

            // transform gradients
            for (size_t i=0; i<gradients.size(); ++i)
                invJacobian.mv(referenceGradients[i][0], gradients[i]);

            // compute matrix entries
            double z = quad[pt].weight() * integrationElement;
            for (int i = 0; i < localSize; ++i)
            {
                int iGlobal = indexSet.subIndex(e, fe.localCoefficients().localKey(i).subEntity(), dim);
                for (int j = i+1; j < localSize; ++j)
                {
                    int jGlobal = indexSet.subIndex(e, fe.localCoefficients().localKey(j).subEntity(), dim);

                    double zij = (gradients[i] * gradients[j]) * z;
                    matrix[iGlobal][jGlobal] += zij;
                    matrix[jGlobal][iGlobal] += zij;
                }
                double zii = (gradients[i] * gradients[i]) * z;
                matrix[iGlobal][iGlobal] += zii;
            }
        }
    }
}


template<class GridView, class Matrix>
void assemblePQ1Mass(const GridView& gridView, Matrix& matrix)
{
    static const int dim = GridView::Grid::dimension;

    typedef typename GridView::IndexSet IndexSet;
    typedef typename GridView::template Codim<0>::Iterator ElementIterator;
    typedef typename GridView::template Codim<0>::Entity Element;
    typedef typename GridView::template Codim<0>::Entity::Geometry Geometry;
    typedef typename Dune::PQkLocalFiniteElementCache<double, double, dim, 1> FiniteElementCache;
    typedef typename FiniteElementCache::FiniteElementType FiniteElement;

    typedef typename Dune::QuadratureRule<double,dim> QuadratureRule;
    typedef typename Dune::template FieldVector<double,dim> LocalCoordinate;
    typedef typename FiniteElement::Traits::LocalBasisType::Traits::RangeType RangeType;

    const IndexSet& indexSet = gridView.indexSet();
    FiniteElementCache cache;

    ElementIterator it = gridView.template begin<0>();
    ElementIterator end = gridView.template end<0>();
    for (; it != end; ++it)
    {
        Element& e = *it;
        const Geometry geometry = e.geometry();
        const FiniteElement& fe = cache.get(e.type());

        int localSize = fe.localBasis().size();

        // get quadrature rule of appropiate order (P1/Q1)
        int order = 0;
        if (e.type().isSimplex())
            order = 2*1;
        else
            order = 2*dim;
        const QuadratureRule& quad = Dune::template QuadratureRules<double, dim>::rule(e.type(), order);

        // store values of shape functions
        std::vector<RangeType> values(localSize);

        for (size_t pt=0; pt < quad.size(); ++pt)
        {
            // get quadrature point
            const LocalCoordinate& quadPos = quad[pt].position();

            // get integration factor
            const double integrationElement = geometry.integrationElement(quadPos);

            // evaluate basis functions
            fe.localBasis().evaluateFunction(quadPos, values);

            // compute matrix entries
            double z = quad[pt].weight() * integrationElement;
            for (int i = 0; i < localSize; ++i)
            {
                int iGlobal = indexSet.subIndex(e, fe.localCoefficients().localKey(i).subEntity(), dim);
                double zi = values[i]*z;
                for (int j = i+1; j < localSize; ++j)
                {
                    int jGlobal = indexSet.subIndex(e, fe.localCoefficients().localKey(j).subEntity(), dim);

                    double zij = values[j] * zi;
                    matrix[iGlobal][jGlobal] += zij;
                    matrix[jGlobal][iGlobal] += zij;
                }
                double zii = values[i] * zi;
                matrix[iGlobal][iGlobal] += zii;
            }
        }
    }
}


template<class GridView, class Vector, class Function>
void assemblePQ1RHS(const GridView& gridView, Vector& r, const Function& f)
{
    static const int dim = GridView::Grid::dimension;

    typedef typename GridView::IndexSet IndexSet;
    typedef typename GridView::template Codim<0>::Iterator ElementIterator;
    typedef typename GridView::template Codim<0>::Entity Element;
    typedef typename GridView::template Codim<0>::Entity::Geometry Geometry;
    typedef typename Dune::PQkLocalFiniteElementCache<double, double, dim, 1> FiniteElementCache;
    typedef typename FiniteElementCache::FiniteElementType FiniteElement;

    typedef typename Dune::QuadratureRule<double,dim> QuadratureRule;
    typedef typename Dune::template FieldVector<double,dim> LocalCoordinate;
    typedef typename FiniteElement::Traits::LocalBasisType::Traits::RangeType RangeType;
    typedef typename Function::RangeType FunctionRangeType;

    const IndexSet& indexSet = gridView.indexSet();
    FiniteElementCache cache;

    ElementIterator it = gridView.template begin<0>();
    ElementIterator end = gridView.template end<0>();
    for (; it != end; ++it)
    {
        Element& e = *it;
        const Geometry geometry = e.geometry();
        const FiniteElement& fe = cache.get(e.type());

        int localSize = fe.localBasis().size();

        // get quadrature rule of appropiate order (P1/Q1)
        int order = 0;
        if (e.type().isSimplex())
            order = 2*1;
        else
            order = 2*dim;
        const QuadratureRule& quad = Dune::template QuadratureRules<double, dim>::rule(e.type(), order);

        // store values of shape functions
        std::vector<RangeType> values(localSize);

        for (size_t pt=0; pt < quad.size(); ++pt)
        {
            // get quadrature point
            const LocalCoordinate& quadPos = quad[pt].position();

            // get integration factor
            const double integrationElement = geometry.integrationElement(quadPos);

            // evaluate basis functions
            fe.localBasis().evaluateFunction(quadPos, values);

            // evaluate function
            FunctionRangeType fAtPos;
            f.evaluate(geometry.global(quadPos), fAtPos);

            // add vector entries
            double z = quad[pt].weight() * integrationElement;
            for (int i = 0; i < localSize; ++i)
            {
                int iGlobal = indexSet.subIndex(e, fe.localCoefficients().localKey(i).subEntity(), dim);

                r[iGlobal].axpy(values[i]*z, fAtPos);
            }
        }
    }
}


template<class Intersection>
bool intersectionContainsVertex(const Intersection& i, int vertexInInsideElement)
{
    static const int dim = Intersection::dimension;

    int faceIdx = i.indexInInside();

    const Dune::ReferenceElement<double,dim>& refElement
        = Dune::ReferenceElements<double, dim>::general(i.inside()->type());

    for (int j = 0; j<refElement.size(faceIdx, 1, dim); ++j)
    {
        int intersectionVertexInElement = refElement.subEntity(faceIdx, 1, j, dim);

        if (intersectionVertexInElement == vertexInInsideElement)
            return true;
    }
    return false;
}


template<class GridView, class BitVector>
void markBoundaryDOFs(const GridView& gridView, BitVector& isBoundary)
{
    static const int dim = GridView::Grid::dimension;

    typedef typename GridView::IndexSet IndexSet;
    typedef typename GridView::template Codim<0>::Iterator ElementIterator;
    typedef typename GridView::template Codim<0>::Entity Element;
    typedef typename GridView::IntersectionIterator IntersectionIterator;
    typedef typename GridView::Intersection Intersection;
    typedef typename Dune::PQkLocalFiniteElementCache<double, double, dim, 1> FiniteElementCache;
    typedef typename FiniteElementCache::FiniteElementType FiniteElement;

    const IndexSet& indexSet = gridView.indexSet();
    FiniteElementCache cache;

    ElementIterator it = gridView.template begin<0>();
    ElementIterator end = gridView.template end<0>();
    for (; it != end; ++it)
    {
        Element& e = *it;
        const FiniteElement& fe = cache.get(e.type());
        int localSize = fe.localBasis().size();

        IntersectionIterator nIt = gridView.ibegin(e);
        IntersectionIterator nEnd = gridView.iend(e);

        for (; nIt!=nEnd; ++nIt)
        {
            const Intersection& i = *nIt;

            if (i.boundary())
            {
                for (int j = 0; j < localSize; ++j)
                {
                    unsigned int vertexInInsideElement = fe.localCoefficients().localKey(j).subEntity();
                    int iGlobal = indexSet.subIndex(e, vertexInInsideElement, dim);
                    if (intersectionContainsVertex(i, vertexInInsideElement))
                        isBoundary[iGlobal] = true;
                }
            }
        }
    }
}


template<class GridType>
GridType* constructCoarseGrid()
{
    const int dim=GridType::dimension;
    typename Dune::GridFactory<GridType> gridFactory;

    typename Dune::FieldVector<double,dim> pos(0);
    gridFactory.insertVertex(pos);
    for (int i=0; i<dim; i++) {
        pos = 0;   pos[i] = 1;
        gridFactory.insertVertex(pos);
    }

    std::vector<unsigned int> element(dim+1);
    for (int i=0; i<dim+1; i++)
        element[i] = i;

    gridFactory.insertElement( Dune::GeometryType(Dune::GeometryType::simplex,dim), element);

    return gridFactory.createGrid();
}


template<class GridType, class TestSuite>
bool checkWithGrid(TestSuite& suite, int uniformRefinements, std::string name)
{
    typename std::template auto_ptr<GridType> gridPtr(constructCoarseGrid<GridType>());
    GridType& grid = *gridPtr;

    grid.globalRefine(uniformRefinements);

    std::cout << "Running test on " << name << " with " << grid.size(0) << " elements." << std::endl;

    bool passed = suite.check(grid);
    if (passed)
        std::cout << "All tests passed with " << name << "." << std::endl;
    return passed;
}


template<class TestSuite>
bool checkWithStandardGrids(TestSuite& suite)
{
    bool passed = true;

#if HAVE_UG
    passed = passed and checkWithGrid<Dune::UGGrid<2> >(suite, 5, "UGGrid<2>");
    passed = passed and checkWithGrid<Dune::UGGrid<3> >(suite, 3, "UGGrid<3>");
#endif

#if HAVE_ALUGRID
    passed = passed and checkWithGrid<Dune::ALUSimplexGrid<2,2> >(suite, 5, "ALUSimplexGrid<2,2>");
//    passed = passed and checkWithGrid<Dune::ALUSimplexGrid<3,3> >(suite, 3, "ALUSimplexGrid<3,3>");
#endif

    return passed;
}

#endif
